@@ -1,10 +1,15 @@
-#!/bin/sh
+#! /bin/bash
 # Check to see if all the needed packages are in the archive.
 # Enrique Zanardi, 1998
 # This is free software under the GNU General Public License.
 
+. ./common.sh
+
+# don't exit on errors for this script
+set +e
+
 if [ $# != 1 ]; then
-echo "Usage: "$0" archive" 1>&2
+echo "Usage: $0 archive" 1>&2
         cat 1>&2 << EOF
 
         archive: the directory containing the Debian binary packages.
@@ -15,20 +20,17 @@ fi
 # Set this to the location of the package archive.
 archive=$1
 
-export arch="`dpkg --print-architecture`"
-
 rootscripts=`pwd`/scripts/rootdisk
 basescripts=`pwd`/scripts/basedisks
-
+extract_subarch_list=`ls $rootscripts/EXTRACT_LIST_${arch}_* 2> /dev/null` || true
 rootfiles="$rootscripts/EXTRACT_LIST_all \
-        $rootscripts/EXTRACT_LIST_${arch}"
+        $rootscripts/EXTRACT_LIST_${arch} \
+	$extract_subarch_list" \
 basefiles="$basescripts/PACKAGES_all \
         $basescripts/PACKAGES_${arch} \
         $basescripts/PARTIAL_PACKAGES_all \
         $basescripts/PARTIAL_PACKAGES_${arch}"
 
-. common.sh
-
 exit=false
 
 for i in $rootfiles $basefiles; do
@@ -44,12 +46,12 @@ fi
 exit=false
 
 # Check to see if all the needed packages are in the archive.
-echo "Checking packages needed in the root disk:"
+echo "checking packages needed in the root disk"
 PACKAGE_PATHS=`make_paths \`cat $rootfiles\``
 if [ $? -ne 0 ]; then
         exit=true;
 fi
-echo "Checking packages needed in the base system:"
+echo "checking packages needed in the base system"
 PACKAGE_PATHS=`make_paths \`cat $basefiles\``
 if [ $? -ne 0 ]; then
         exit=true;
@@ -58,4 +60,3 @@ fi
 if $exit; then
 	exit -1
 fi
-
