@@ -2,9 +2,7 @@
 # Low mem initrd Disk maker. Enrique Zanardi, June 1998
 # This is free software under the GNU General Public License.
 
-set -e
-set -v
-set -x
+. ./common.sh
 
 export LANG=C
 
@@ -16,16 +14,16 @@ fi
 PROGRAM=utilities/lowmemrd/linuxrc
 
 # Set this to the block device for the floppy disk.
-floppy=/var/tmp/floppy-image.$$
+floppy=$(tempfile -p fimg)
 
-mnt=/var/tmp/mnt.lowmemrd
+mnt=${tmpdir}/boot-floppies/mnt.lowmemrd
 
 (cd utilities/lowmemrd; make)
 
 # Try to unmount anything that should not be mounted.
 (umount $mnt; umount $floppy; true)2>/dev/null >/dev/null
 
-mkdir -p $mnt
+make_tmpdir $mnt
 
 blocks=`/bin/ls -l $PROGRAM |awk '{print $5}'`
 blocks=`expr ${blocks} / 1024 + 30`
@@ -41,7 +39,7 @@ dd if=/dev/zero of=$floppy bs=1k count=${blocks}
 mount -t minix -o loop $floppy $mnt
 
 mkdir $mnt/dev
-(cd $mnt/dev; /dev/MAKEDEV fd0 fd1 hda hdb hdc hdd sda sdb sdc sdd)
+(cd $mnt/dev; /sbin/MAKEDEV fd0 fd1 hda hdb hdc hdd sda sdb sdc sdd)
 mknod $mnt/dev/tty1 c 4 1
 mkdir $mnt/proc
 mkdir $mnt/mnt
