@@ -1,7 +1,15 @@
+
+/* to avoid parse error before `__extension__' in <linux/string.h> included by
+ * <linux/genhd.h> (conflicts with <bits/string2.h> on sparc at least).
+ */
+#define __NO_STRING_INLINES
+
+#include <ctype.h>
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
 #include <stdlib.h>
+#include <sys/stat.h>
 #include <unistd.h>
 #include <linux/unistd.h>
 #include <fcntl.h>
@@ -10,12 +18,23 @@
 #include <linux/fs.h>
 #include <mntent.h>
 #include <assert.h>
+#include <syslog.h>
 #include <linux/cdrom.h>
+#include <linux/major.h>
+#include <linux/genhd.h>
 
 #include "fdisk.h"
 
 #define SIZE(a) (sizeof(a)/sizeof(a[0]))
 
+/*
+ * XXX: Some 2.2.x kernel headers (ie. libc6-dev) do
+ * not contain this for some reason -- Ben
+ */
+#ifndef MAX_DISKNAME_LEN
+#define MAX_DISKNAME_LEN 32
+#endif
+
 /* this array translates partition type numbers to file system types */
 static struct fstypes {
     int ptype;
@@ -30,9 +49,6 @@ static struct fstypes {
     { PTYPE_WFAT32,		FSTYPE_MSDOS },
     { PTYPE_WFAT32L,		FSTYPE_MSDOS },
     { PTYPE_WEXT,		FSTYPE_EXTPART },
-    { PTYPE_MINIX_DR,		FSTYPE_MINIX },
-    { PTYPE_SWAP_DR,		FSTYPE_SWAP },
-    { PTYPE_LINUX_DR,		FSTYPE_EXT2 },
     { PTYPE_MINIX,		FSTYPE_MINIX },
     { PTYPE_LSWAP,		FSTYPE_SWAP },
     { PTYPE_LINUX,		FSTYPE_EXT2 },
@@ -102,6 +118,14 @@ static struct fstypes {
      * make them the transition easier :-) */
     { PTYPE_MAC_AUX,		FSTYPE_EXT2 },
 #endif /* HAVE_MAC_PARTITION */
+#if HAVE_ACORN_PARTITION
+    { PTYPE_ACORN_ADFS,		FSTYPE_ADFS },
+    { PTYPE_ACORN_EXT2,		FSTYPE_EXT2 },
+    { PTYPE_ACORN_SWAP,		FSTYPE_SWAP },
+    { PTYPE_ACORN_ICS_ADFS,	FSTYPE_ADFS },
+    { PTYPE_ACORN_ICS_EXT2,	FSTYPE_EXT2 },
+    { PTYPE_ACORN_ICS_SWAP,	FSTYPE_SWAP },
+#endif /* HAVE_ACORN_PARTITION */
 #if HAVE_LOOP_PARTITION
     /* Fake partition types for loop mounts */
     { PTYPE_LOOP_EXT2,		FSTYPE_EXT2 },
@@ -197,6 +221,7 @@ static struct systypes {
     {0xf1, "SpeedStor"},
     {0xf2, "DOS 3.3+ secondary"},
     {0xf4, "SpeedStor large partition"},
+    {0xfd, "Linux RAID autodetect"},
     {0xfe, "SpeedStor >1024 cyl. or LANstep"},
     {0xff, "Xenix Bad Block Table"},
 #endif /* HAVE_MSDOS_PARTITION */
@@ -236,6 +261,7 @@ static struct systypes {
     { PTYPE_OSF_BOOT,	"boot" },
     { PTYPE_OSF_ADOS,	"Amiga AFFS" },
     { PTYPE_OSF_HFS,	"Apple HFS" },
+    { PTYPE_OSF_ADVFS,	"Digital AdvFS" },
 #endif /* HAVE_OSF_PARTITION */
 #if HAVE_SUN_PARTITION
     { PTYPE_SUN_EMPTY,		"empty" },
@@ -307,6 +333,15 @@ static struct systypes {
     { PTYPE_MAC_AFFS,		"Amiga AFFS" },
     { PTYPE_MAC_EXT2,		"Linux native" },
 #endif /* HAVE_MAC_PARTITION */
+#if HAVE_ACORN_PARTITION
+    { PTYPE_ACORN_LINUX,	"ARM Linux Partition Map" },
+    { PTYPE_ACORN_ADFS,		"ADFS" },
+    { PTYPE_ACORN_EXT2,		"Linux ext2 (ARM Linux)" },
+    { PTYPE_ACORN_SWAP,		"Linux swap (ARM Linux)" },
+    { PTYPE_ACORN_ICS_ADFS,	"ADFS (ICS/APDL)" },
+    { PTYPE_ACORN_ICS_EXT2,	"Linux ext2 (ICS/APDL)" },
+    { PTYPE_ACORN_ICS_SWAP,	"Linux swap (ICS/APDL)" },
+#endif /* HAVE_ACORN_PARTITION */
 #if HAVE_LOOP_PARTITION 
     { PTYPE_LOOP_EXT2,		"Linux native (loop mounted)" },
     { PTYPE_LOOP_SWAP,		"Linux swap (loop mounted)" },
@@ -326,6 +361,7 @@ char *fstype_name[FSTYPE_MAX] = {
     "minix",
     "affs",
     "hfs",
+    "adfs",
     "nfs",
     NULL, /* FSTYPE_EXTPART is not a valid filesystem, so return "UNKNOWN" */
     NULL /* FSTYPE_UNKNOWN */
@@ -335,7 +371,7 @@ char *fstype_name[FSTYPE_MAX] = {
 
 /* translate partition type number to description */
 char *
-sysname(int type) {
+fdisk_sysname(int type) {
     struct systypes *s;
 
     for( s = sys_types+SIZE(sys_types)-1; s >= sys_types; s-- )
@@ -379,8 +415,6 @@ void fdisk_init() {
 
   fdisk_disks = last_disk = NULL;
   fdisk_partitions = last_partition = NULL;
-  mounted_partitions = NULL;
-  swapon_partitions = last_swapon_partition = NULL;
 }
 
 
@@ -396,6 +430,7 @@ fdisk_add_disk(const char *name, unsigned long size){
   disk->next = NULL;
   disk->partitions = NULL;
   disk->last_partition = NULL;
+  disk->partition_format = NULL;
 
   if (NULL == fdisk_disks) {
     fdisk_disks = disk;
@@ -433,63 +468,191 @@ fdisk_fstype_name_of(unsigned int ptype)
     return fstype_name[fstype];
 }
 
-struct fdisk_partition *
-fdisk_add_partition(const char *name, int minor, unsigned int type,
-		    unsigned long size){
 
-  struct fdisk_partition *partition;
-  struct fdisk_disk *disk;
-  int fstype;
+/* get_part_info: read and parse line in /proc/partitions format */
+static int get_part_info(FILE *f, dev_t *dev, char *name)
+{
+    char line[64];
+    unsigned int major, minor;
 
-  partition = (struct fdisk_partition *) 
-    malloc(sizeof(struct fdisk_partition));
-  if (!partition) return NULL;
+    while (fgets(line, sizeof line, f)) {
+	if (sscanf(line, "%4u  %4u %*u %s", &major, &minor, name) == 3) {
+	    *dev = MKDEV(major, minor);
+	    return 1;
+	}
+    }
+    return 0;
+}
 
-  if (minor != -1){
-    partition->name= malloc(strlen(name)+4);
-    sprintf(partition->name,"%s%d", name, minor);
-  } else {
-    partition->name=strdup(name);
-  }
-  partition->mount_point= NULL;
-  partition->type = type;
-  partition->size = size;
-  partition->in_use = 0;
-  partition->next = NULL;
-  partition->next_by_disk = NULL;
-  partition->next_by_type = NULL;
-  partition->next_in_use = NULL;
-  
-  if (minor != -1){
-    disk=fdisk_find_disk(name);
-#ifdef DEBUG
-    printf( "adding partition %s, to disk %s\n", partition->name, disk->name );
-#endif
-    if (NULL == disk->partitions)
-      disk->partitions = partition;
-    if (NULL != disk->last_partition)
-      (disk->last_partition)->next_by_disk = partition;
-    disk->last_partition = partition;
-  }
-  
-  if (NULL == fdisk_partitions)
-    fdisk_partitions = partition;
-  if (NULL != last_partition)
-    last_partition->next = partition;
-  last_partition = partition;
 
-  fstype = fdisk_fstype_of(type);
-  if (NULL == fdisk_partitions_by_type[fstype])
-    fdisk_partitions_by_type[fstype] = partition;
-  if (NULL != last_partitions_by_type[fstype])
-    last_partitions_by_type[fstype]->next_by_type = partition;
-  last_partitions_by_type[fstype] = partition;
+static struct DEVICE_BLOCK {
+    dev_t dev;	/* initial device number */
+    int n_dev;	/* number of drives in this block */
+    int step;	/* step in device number between drives */
+} device_blocks[] = {
+    /* all the r/w block devices known to the kernel */
+    /* according to linux/Documentation/devices.txt  */
+    {  3 << 8,   2, 64 },	/* hda      - hdb       */
+    {  8 << 8,  16, 16 },	/* sda      - sdp       */
+    { 13 << 8,   2, 64 },	/* xda      - xhb       */
+    { 14 << 8,   4, 64 },	/* dos_hda  - dos_hdd   */
+    { 21 << 8,   2, 64 },	/* mfma     - mfmb      */
+    { 22 << 8,   2, 64 },	/* hdc      - hdd       */
+    { 28 << 8,  16, 16 },	/* ada      - adp       */
+    { 33 << 8,   2, 64 },	/* hde      - hdf       */
+    { 34 << 8,   2, 64 },	/* hdg      - hdh       */
+    { 36 << 8,   4, 64 },	/* eda      - edd       */
+    { 44 << 8,  16, 16 },	/* ftla     - ftlp      */
+    { 45 << 8,   4, 16 },	/* pdaa     - pdad      */
+    { 48 << 8, 256,  8 },	/* rd/c0d0  - rd/c8d31  */
+    { 56 << 8,   2, 64 },	/* hdi      - hdj       */
+    { 57 << 8,   2, 64 },	/* hdk      - hdl       */
+    { 65 << 8, 112, 16 },	/* sdq      - sddx      */
+    { 72 << 8, 128, 16 },	/* ida/c0d0 - ida/c7d15 */
+    {       0,   0,  0 }
+};
+
+
+/* check_part_offset: return true iff partition number is within range */
+static int check_part_offset(dev_t disk, unsigned int part)
+{
+    struct DEVICE_BLOCK *d;
+    int i;
+
+    for (d = device_blocks; d->dev && disk >= d->dev; d++)
+	for (i = 0; i < d->n_dev; i++)
+	    if (disk == d->dev + i * d->step)
+		return part < d->step;
+
+    return 0;
+}
+
+
+/* strdupcat: return concatenation of two strings in malloc-ed buffer */
+static char *strdupcat(char *s1, char *s2)
+{
+    char *out = malloc(strlen(s1) + strlen(s2) + 1);
+    if (out) {
+	strcpy(out, s1);
+	strcat(out, s2);
+    }
+    return out;
+}
+
+
+/* part_name: return partition device name from disk and partition number */
+static char *part_name(struct fdisk_disk *disk, unsigned int part)
+{
+    FILE *f;
+    char *pname = NULL, *dname = disk->name + 5;
+
+    assert(strncmp(disk->name, "/dev/", 5) == 0);
 
 #ifdef DEBUG
-  printf( "added partition %s, size %ld, type %08x=\"%s\"\n",
-	  partition->name, size, type, sysname(type) );
+    printf("looking for partition %d of disk %s\n", part, dname );
 #endif
-  return(partition);
+
+    /* scan /proc/partitions for likely devices */
+    f = fopen("/proc/partitions", "r");
+    if (f) {
+	char name[MAX_DISKNAME_LEN];
+	dev_t dev, disk_dev = 0;
+
+	while (get_part_info(f, &dev, name)) {
+	    /* look for target partition or real disk containing it */
+	    if (! disk_dev) {
+		/* look for real disk containing target partition */
+		if (strcmp(name, dname) == 0) {
+		    if (! check_part_offset(dev, part)) {
+			syslog(LOG_WARNING, "disk %s partition %u is not "
+		               "supported (no such device)", dname, part);
+			break;
+		    }
+		    disk_dev = dev;
+		}
+	    } else {
+		/* look for partition with same number as target */
+		if (disk_dev + part == dev) {
+#ifdef DEBUG
+		    printf("found partition %s: major=%d, minor=%d\n",
+			   name, (int)MAJOR(dev), (int)MINOR(dev));
+#endif
+		    pname = strdupcat("/dev/", name);
+		    break;
+		}
+	    }
+	}
+	fclose(f);
+    }
+    return pname;
+}
+
+
+struct fdisk_partition *fdisk_add_partition(const char *name, int minor,
+ unsigned int type, unsigned long size)
+{
+    int fstype;
+    struct fdisk_partition *partition = malloc(sizeof(struct fdisk_partition));
+
+    if (!partition)
+        return NULL;
+
+    partition->name = NULL;
+    partition->mount_point= NULL;
+    partition->type = type;
+    partition->size = size;
+    partition->in_use = 0;
+    partition->disk = NULL;
+    partition->next = NULL;
+    partition->next_by_disk = NULL;
+    partition->next_by_type = NULL;
+    partition->next_in_use = NULL;
+
+    if (minor == -1) {
+        partition->name = strdup(name);
+        assert(partition->name);
+    } else {
+	partition->disk = fdisk_find_disk(name);
+	assert(partition->disk);
+	assert(minor >= 0);
+	partition->name = part_name(partition->disk, (unsigned) minor);
+	if (!partition->name) {
+	    syslog(LOG_WARNING, "no device name for disk %s partition %d "
+		   "(skipped)", name, minor);
+	    free(partition);
+	    return NULL;
+	}
+
+#ifdef DEBUG
+        printf("adding partition %s, to disk %s\n",
+               partition->name, partition->disk->name);
+#endif
+
+        if (! partition->disk->partitions)
+            partition->disk->partitions = partition;
+        if (partition->disk->last_partition)
+            partition->disk->last_partition->next_by_disk = partition;
+        partition->disk->last_partition = partition;
+    }
+  
+    if (NULL == fdisk_partitions)
+        fdisk_partitions = partition;
+    if (NULL != last_partition)
+        last_partition->next = partition;
+    last_partition = partition;
+
+    fstype = fdisk_fstype_of(type);
+    if (NULL == fdisk_partitions_by_type[fstype])
+        fdisk_partitions_by_type[fstype] = partition;
+    if (NULL != last_partitions_by_type[fstype])
+        last_partitions_by_type[fstype]->next_by_type = partition;
+    last_partitions_by_type[fstype] = partition;
+
+#ifdef DEBUG
+    printf("added partition %s, size %ld, type %08x=\"%s\"\n",
+	   partition->name, size, type, fdisk_sysname(type) );
+#endif
+    return partition;
 }
 
 
@@ -544,6 +707,21 @@ fdisk_find_partition_by_name(const char *name){
   return(NULL);
 }
 
+struct fdisk_partition *
+fdisk_find_partition_by_type(const int type){
+  struct fdisk_partition *p;
+
+  p = fdisk_partitions;
+
+  while (p) {
+    if (p->type == type)
+      return(p);
+    p = p->next;
+  }
+
+  return(NULL);
+} 
+
 void
 mounted_reread(){
   FILE *mtabFile;
@@ -551,7 +729,9 @@ mounted_reread(){
   struct fdisk_partition *p; 
   int type;
   
-  if (NULL == (mtabFile = setmntent("/etc/mtab", "r"))) {
+  mounted_partitions = NULL;
+
+  if (NULL == (mtabFile = setmntent("/proc/mounts", "r"))) {
     return ;
   }
 
@@ -560,10 +740,11 @@ mounted_reread(){
     fprintf(stderr,"mounted partition: %s, on: %s, type: %s , options: %s ... ", 
 	mnt_ent->mnt_fsname, mnt_ent->mnt_dir, mnt_ent->mnt_type, mnt_ent->mnt_opts);
 #endif
-    /* If mnt_ent is proc, ignore it */
-    if (strcmp(mnt_ent->mnt_fsname,"proc") == 0) {
+    /* If mnt_ent is proc or devpts, ignore it */
+    if ( (strcmp(mnt_ent->mnt_type,"proc") == 0) ||
+	 (strcmp(mnt_ent->mnt_type,"devpts") == 0) ) {
 #ifdef DEBUG
-      fprintf(stderr,"proc fs. Ignoring it.\n");
+      fprintf(stderr,"proc/devpts fs. Ignoring it.\n");
 #endif
       continue;
     }
@@ -631,12 +812,17 @@ mounted_reread(){
   endmntent(mtabFile);
 }
 
+#ifdef OLDKERNEL	/* For 2.0.x kernels */
+
 void
 swapon_reread(){
   FILE *swapsFile;
   struct mntent *swap_ent;
+
   struct fdisk_partition *p;
 
+  swapon_partitions = last_swapon_partition = NULL;
+
   if (NULL == (swapsFile = setmntent("/etc/swaps", "r"))) {
     return ;
   }
@@ -669,6 +855,101 @@ swapon_reread(){
   endmntent(swapsFile);
 }
 
+#else /* For 2.1.x kernels */
+
+struct swpent
+  {
+    char *name;
+    int type; /* 0 = partition, 1 = file */
+    int size;
+    int used;
+    int prio;
+  };
+
+struct swpent *
+getswpent (FILE *stream, struct swpent *sp)
+{
+  char *head;
+#define BUFSIZE 8192
+  char buffer[BUFSIZE];
+
+  do
+    {
+      char *end_ptr;
+
+      if (fgets (buffer, BUFSIZE, stream) == NULL)
+        return NULL;
+
+      end_ptr = strchr (buffer, '\n');
+      if (end_ptr != NULL)      /* chop newline */
+        *end_ptr = '\0';
+      else
+        {
+          /* Not the whole line was read.  Do it now but forget it.  */
+          char tmp[1024];
+          while (fgets (tmp, sizeof tmp, stream) != NULL)
+            if (strchr (tmp, '\n') != NULL)
+              break;
+        }
+      head = buffer;
+      /* skip empty lines and comment lines:  */
+    } while (head[0] != '/');
+
+  head += strcspn (head, " \t");
+  *head++='\0';
+  sp->name = strdup(buffer);
+  head += strspn (head, " \t");
+  sp->type = (strncmp(head, "file", 4) == 0) ? 1 : 0;
+  head += strcspn (head, " \t");
+  sscanf(head," %d %d %d ", &sp->size, &sp->used, &sp->prio);
+
+  return sp;
+}
+
+void
+swapon_reread(){
+  FILE *swapsFile;
+  struct swpent swap_ent;
+
+  struct fdisk_partition *p;
+
+  swapon_partitions = last_swapon_partition = NULL;
+
+  if (NULL == (swapsFile = setmntent("/proc/swaps", "r"))) {
+    return ;
+  }
+
+  while ( NULL != getswpent(swapsFile,&swap_ent) ) {
+#ifdef DEBUG
+    fprintf(stderr,"Swap partition: %s, type: %s , size: %d, used: %d, prio: %d ... ",
+	swap_ent.name, (swap_ent.type==0)?"partition":"file", swap_ent.size,
+	swap_ent.used, swap_ent.prio);
+#endif
+    /* If swap_ent is not a partition on the local disks, then ignore it */
+    p = fdisk_find_partition_by_name(swap_ent.name);
+    free(swap_ent.name);
+    if ( NULL != p) {
+#ifdef DEBUG
+      fprintf(stderr,"found on local disks.\n");
+#endif
+      p->in_use=-1;
+      if (NULL == swapon_partitions)
+        swapon_partitions=p;
+      else {
+        last_swapon_partition->next_in_use=p;
+      }
+      last_swapon_partition=p;
+    }
+#ifdef DEBUG
+    else {
+      fprintf(stderr,"NOT found on local disks. Ignored.\n");
+    }
+#endif
+  }
+  fclose(swapsFile);
+}
+#endif /* OLDKERNEL */
+
 /*
  * sseek: seek to specified sector - return 0 on failure
  *
@@ -705,7 +986,7 @@ static int sseek(int fd, unsigned long s)
     return 1;
 }
 #else
-#ifndef __alpha__
+#if !defined(__alpha__) && !defined(__ia64__)
 /* Better print an error if _llseek isn't available at compile time, the
  * resulting binary wouldn't be able to handle disks > 2GB */
 #error No _llseek available
@@ -745,7 +1026,8 @@ static struct {
     { "sun",   parse_sun_partition },
     { "amiga", parse_amiga_partition },
     { "atari", parse_atari_partition },
-    { "mac",   parse_mac_partition }
+    { "mac",   parse_mac_partition },
+    { "acorn", parse_acorn_partition }
 };
 
 static void
@@ -766,73 +1048,151 @@ parse_partition(char *device, int fd)
     }
 }
 
-static struct trylist {
-    char *dev_pattern;
-    char start, end;
-} trylist[] = {
-    { "/dev/hd", 'a', 'd' },
-    { "/dev/sd", 'a', 'h' },
-#if #cpu (i386)
-    { "/dev/ed", 'a', 'd' },
-#endif
-#if #cpu (m68k)
-    { "/dev/ad", 'a', 'h' },
-#endif
-/* 
- * #ifdef DEBUG
- *     { "/dev/loop", '0', '7' },
- * #endif
- */
-};
 
-static int timed_out;
-
-static void alarmed(int signo)
+/* ide_is_disk: return true if IDE device is a disk drive */
+static int ide_is_disk(char *device)
 {
-    timed_out = 1;
+    char pname[32];
+    int n;
+    FILE *f;
+    int result = 0;
+
+    /* /dev/hd? => /proc/ide/hd?/media */
+    device = strrchr(device, '/');
+    assert(device);
+    n = snprintf(pname, sizeof pname, "/proc/ide/%s/media", device + 1);
+    assert(n > 0);
+
+    /* open /proc/ide/hd?/media and look for ide_disk text */
+    f = fopen(pname, "r");
+    if (f) {
+	char line[8];
+	result = fgets(line, sizeof line, f) && strcmp(line, "disk\n") == 0;
+	fclose(f);
+    }
+    return result;
 }
 
-void
-fdisk_reread(void)
-{
-    struct trylist *p;
-    struct cdrom_volctrl cdvol;
-    int i, fd;
-    char letter, device[12], *q;
-    unsigned size;
-    
-    fdisk_init();
-    
-    for( p = trylist, i = 0; i < SIZE(trylist); ++p, ++i ) {
-	strcpy( device, p->dev_pattern );
-	q = device+strlen(device);
-	q[1] = 0;
-	for( letter = p->start; letter <= p->end; ++letter ) {
-	    *q = letter;
-	    signal( SIGALRM, alarmed );
-	    timed_out = 0;
-	    alarm( 30 );
-	    if ((fd = open( device, O_RDONLY )) >= 0 && !timed_out) {
-		alarm( 0 );
-		if (ioctl(fd, CDROMVOLREAD, &cdvol )!= 0) {
-      /* if returns EINVAL or EPERM or EIO, this is not a CD-ROM drive */
-      		    if ((errno==EINVAL)||(errno==EPERM)||(errno==EIO)) {
-			if (ioctl(fd, BLKGETSIZE, &size))
-			    size = 0; /* don't complain, size not really 
-					 used yet */
-			size >>= 1; /* convert from 512-byte sectors to kB */
-			fdisk_add_disk( device, size );
 
-			parse_partition( device, fd );
-		    }
-		}
-		close( fd );
-	    }
-	    else {
-		/* else: don't complain, ignore open errors */
-		alarm( 0 );
+/* is_disk: return true if device is a disk drive */
+static int is_disk(char *device, int fd)
+{
+    struct stat stat_buf;
+    struct cdrom_volctrl cdvol;
+
+    /* reject anything un-stat-able, or not a block device */
+    if (fstat(fd, &stat_buf) || ! S_ISBLK(stat_buf.st_mode))
+	return 0;
+
+    /* IDE devices */
+    switch (MAJOR(stat_buf.st_rdev)) {
+    case IDE0_MAJOR: case IDE1_MAJOR: case IDE2_MAJOR:
+    case IDE3_MAJOR: case IDE4_MAJOR: case IDE5_MAJOR:
+	return ide_is_disk(device);
+    }
+
+    /* SCSI disk driver will only accept disks and mag-opt drives */
+    if (SCSI_DISK_MAJOR(MAJOR(stat_buf.st_rdev))) 
+	return 1;
+
+#if #cpu (i386)
+    /* ESDI devices assumed to all be disks */
+    if (MAJOR(stat_buf.st_rdev) == PS2ESDI_MAJOR)
+	return 1;
+#elif #cpu (m68k)
+    /* ACSI devices assumed to all be disks */
+    if (MAJOR(stat_buf.st_rdev) == ACSI_MAJOR)
+	return 1;
+#endif
+
+    /** TODO ***************************************************************/
+    /* I'm not sure how to check for non-disk devices on dac960 or smart2. */
+    /* *********************************************************************/
+
+    /* supposedly anything failing CDROMVOLREAD for the right reasons is ok */
+    return ioctl(fd, CDROMVOLREAD, &cdvol) && (errno == EINVAL ||
+					       errno == EPERM  ||
+					       errno == EIO    ||
+					       errno == EBADRQC);
+}
+
+
+static void alarmed(int signo) { }
+
+
+/* try_device: check specific device by name, add to disk/part lists if ok */
+static void try_device(char *device)
+{
+    int fd;
+
+    signal(SIGALRM, alarmed);
+    alarm(30);
+    fd = open(device, O_RDONLY);
+    alarm(0);
+
+    if (fd >= 0) {
+        if (is_disk(device, fd)) {
+	    unsigned int size;
+	    if (ioctl(fd, BLKGETSIZE, &size))
+		size = 0;   /* don't complain, size not really used yet */
+	    size >>= 1;     /* convert from 512-byte sectors to kB */
+	    fdisk_add_disk(device, size);
+
+	    parse_partition(device, fd);
+	} else {
+	    /* Don't need to do anything for individual partitions because   */
+	    /* parse_partition() above calls fdisk_add_partition() for every */
+	    /* partition on the 'whole-disk' device.  We could eliminate a   */
+	    /* huge amount of code and complexity if we knew the partition   */
+	    /* type right now.						     */
+	}
+	close(fd);
+    }
+}
+
+
+/* is_drive: return true if device is a 'whole disk', not just a partition */
+static int is_drive(dev_t dev)
+{
+    struct DEVICE_BLOCK *d;
+    int i;
+
+    /* search for device in each block of devices */
+    for (d = device_blocks; d->dev && dev >= d->dev; d++)
+	for (i = 0; i < d->n_dev; i++)
+	    if (dev == d->dev + i * d->step)
+		return 1;
+
+    return 0;
+}
+
+
+/* fdisk_reread: build up disk/partition lists from scratch */
+void fdisk_reread(void)
+{
+    FILE *f;
+
+    fdisk_init();
+
+    /* scan /proc/partitions for likely device names */
+    f = fopen("/proc/partitions", "r");
+    if (f) {
+	char name[MAX_DISKNAME_LEN];
+	dev_t dev;
+
+	while (get_part_info(f, &dev, name)) {
+	    /* check device number against known 'whole-disk' devices */
+	    if (is_drive(dev)) {
+		static char path[] = "/dev/";
+		char dev_name[sizeof path + sizeof name];
+
+		/* try to build partition list for device */
+		strcpy(dev_name, path);
+		strcat(dev_name, name);
+		try_device(dev_name);
 	    }
 	}
+	fclose(f);
     }
 
     mounted_reread();
