@@ -52,6 +52,10 @@
  *
  */
 
+/* to avoid parse error before `__extension__' in <linux/string.h> included by
+ * <linux/genhd.h> (conflicts with <bits/string2.h> on sparc at least).
+ */
+#define __NO_STRING_INLINES
 
 #include <unistd.h>
 #include <stdio.h>
@@ -186,7 +190,7 @@ struct systypes sys_types[] = { /* struct systypes in fdisk.h *//* bf */
 		{0xff, "BBT"}			/* (bad track table) */
 	};
 
-int nsys_types = sizeof (sys_types) / sizeof (struct systypes); /* bf */
+#define nsys_types (int)(sizeof (sys_types) / sizeof (struct systypes)) /* bf */
 
 jmp_buf listingbuf;
 
@@ -267,8 +271,7 @@ void xmenu(void)
 
 char *partition_type(unsigned char type)
 {
-	int high = sizeof(sys_types) / sizeof(struct systypes),
-		low = 0, mid;
+	int high = nsys_types, low = 0, mid;
 	uint tmp;
 
 	while (high >= low) {
